///////////////////////////////////////////////////////////////////////////////
// ketchup/message_processor.hpp
//
//  Version 0.904
//  Copyright 2005 MB@unknown.jp.
//
// This file is a part of the *ketchup*.
// The use and distribution terms for this software are covered by the
// Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
// which can be found in the file CPL.TXT at the root of this distribution.
// By using this software in any fashion, you are agreeing to be bound by
// the terms of this license. You must not remove this notice, or
// any other, from this software.
//
//  Revision History
//   05 May 2005 Initial version

///////////////////////////////////////////////////////////////////////////////
// Concepts:
//
//  MessageMap is...
//   a MPL Sequence of IfStatement.
//

#ifndef KETCHUP_MESSAGE_PROCESSOR_H
#define KETCHUP_MESSAGE_PROCESSOR_H

#ifdef _MSC_VER
#pragma once
#endif

#include <boost/call_traits.hpp>
#include <boost/concept_check.hpp>
#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/mpl/apply.hpp>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/eval_if.hpp>
#include <boost/mpl/is_sequence.hpp>
#include <boost/mpl/placeholders.hpp>
#include <boost/spirit/fusion/sequence/tuple.hpp>
#include <boost/spirit/fusion/sequence/get.hpp>

#include "big_if_statement.hpp"
#include "ignore_unused_variables_warning.hpp"

namespace ketchup {

	namespace detail {

	template< class Derived >
	struct process_window_message_params
	{
		typedef boost::fusion::tuple<
			Derived&, HWND, UINT, WPARAM, LPARAM, LRESULT&, DWORD, BOOL&
		> type;
	};

	template< class Derived >
	struct process_window_message_params_reference
	{
		typedef typename process_window_message_params<Derived>::type params_t;
		typedef typename boost::call_traits<params_t>::reference type;
	};

	template< class Derived >
	struct process_window_message_params_param_type
	{
		typedef typename process_window_message_params<Derived>::type params_t;
		typedef typename boost::call_traits<params_t>::param_type type;
	};


	///////////////////////////////////////////////////////////////////////////////
	// message_processor_base
	//
	template< class Derived, class HasCrackedHandlers >
	struct message_processor_base
	{
		typedef typename process_window_message_params<Derived>::type
			process_window_message_params_t;
		typedef typename process_window_message_params_reference<Derived>::type
			process_window_message_params_t_reference;

		struct base
		{
			template< class MessageMap >
			bool process_window_message(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT& lResult, DWORD dwMsgMapID)
			{
				Derived& derived = *static_cast<Derived*>(this);
				BOOL bHandled = TRUE;
				process_window_message_params_t params(derived, hWnd, uMsg, wParam, lParam, lResult, dwMsgMapID, bHandled);

				typedef typename big_if_statement<MessageMap, process_window_message_params_t_reference>::type body;
				return body::compile(params);
			};

			typedef base type;
		};	

		struct base_for_cracked_handlers
		{
			bool b_msg_handled;
			bool is_msg_handled() { return b_msg_handled; }
			void set_msg_handled(bool handled) { b_msg_handled = handled; }

			template< class MessageMap >
			bool process_window_message(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT& lResult, DWORD dwMsgMapID)
			{
				Derived& derived = *static_cast<Derived*>(this);
				BOOL bHandled = TRUE;
				process_window_message_params_t params(derived, hWnd, uMsg, wParam, lParam, lResult, dwMsgMapID, bHandled);

				bool old = derived.is_msg_handled();	// for recursive calls
				typedef typename big_if_statement<MessageMap, process_window_message_params_t_reference>::type body;
				bool ret = body::compile(params);
   			derived.set_msg_handled(old);
				return ret;
			};

			typedef base_for_cracked_handlers type;
		};

		typedef typename boost::mpl::eval_if<HasCrackedHandlers,
			base_for_cracked_handlers,
			base
		>::type type;
	};

	} // namespace detail 


///////////////////////////////////////////////////////////////////////////////
// message_processor
//
	template< class Derived, bool has_cracked_handlers = false >
	struct message_processor :
		detail::message_processor_base< Derived, boost::mpl::bool_<has_cracked_handlers> >::type
	{
	protected:
		typedef typename detail::process_window_message_params<Derived>::type
			process_window_message_params_t;
		typedef typename detail::process_window_message_params_reference<Derived>::type
			process_window_message_params_t_reference;
		typedef typename detail::process_window_message_params_param_type<Derived>::type
			process_window_message_params_t_param_type;

	///////////////////////////////////////////////////////////////////////////////
	// shortcuts
	//
	static Derived&	get_derived(process_window_message_params_t_param_type params)
	{
		return boost::fusion::get<0>(params);
	}

	static HWND get_hWnd(process_window_message_params_t_param_type params)
	{
		return boost::fusion::get<1>(params);
	}

	static UINT get_uMsg(process_window_message_params_t_param_type params)
	{
		return boost::fusion::get<2>(params);
	}

	static WPARAM get_wParam(process_window_message_params_t_param_type params)
	{
		return boost::fusion::get<3>(params);
	}

	static LPARAM get_lParam(process_window_message_params_t_param_type params)
	{
		return boost::fusion::get<4>(params);
	}

	static LRESULT& get_lResult(process_window_message_params_t_param_type params)
	{
		return boost::fusion::get<5>(params);
	}

	static DWORD get_dwMsgMapID(process_window_message_params_t_param_type params)
	{
		return boost::fusion::get<6>(params);
	}

	static BOOL& get_bHandled(process_window_message_params_t_param_type params)
	{
		return boost::fusion::get<7>(params);
	}

	///////////////////////////////////////////////////////////////////////////////
	// IfStatements of MessageMap
	//
	// Note: Default parameters in nested templates will be ignored by VC7.1.
	//

	///////////////////////////////////////////////////////////////////////////////
	// empty_handler
	//
	struct empty_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			boost::ignore_unused_variable_warning(params);
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// message_handler
	//
	typedef LRESULT (Derived::*message_handler_type)(UINT, WPARAM, LPARAM, BOOL&);

	template< UINT msg, message_handler_type func >
	struct message_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == msg) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)(get_uMsg(params), get_wParam(params), get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	template< UINT msg, message_handler_type func >
	struct message_handler_hot
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == msg) {
				get_lResult(params) = (get_derived(params).*func)(get_uMsg(params), get_wParam(params), get_lParam(params), get_bHandled(params));
				return true;
			}
			return false;
		}
	};

	template< UINT msg, message_handler_type func >
	struct message_handler_not
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == msg) {
				get_lResult(params) = (get_derived(params).*func)(get_uMsg(params), get_wParam(params), get_lParam(params), get_bHandled(params));
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// message_range_handler
	//
	template< UINT msgFirst, UINT msgLast, message_handler_type func >
	struct message_range_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) >= msgFirst && get_uMsg(params) <= msgLast) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)(get_uMsg(params), get_wParam(params), get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_handler
	//
	typedef LRESULT (Derived::*command_handler_type)(WORD, WORD, HWND, BOOL&);

	template< WORD id, WORD code, command_handler_type func >
	struct command_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == WM_COMMAND && id == LOWORD(get_wParam(params)) && code == HIWORD(get_wParam(params))) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)(HIWORD(get_wParam(params)), LOWORD(get_wParam(params)), (HWND)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_id_handler
	//
	template< WORD id, command_handler_type func >
	struct command_id_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if(get_uMsg(params) == WM_COMMAND && id == LOWORD(get_wParam(params))) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)(HIWORD(get_wParam(params)), LOWORD(get_wParam(params)), (HWND)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	template< WORD id, command_handler_type func >
	struct command_id_handler_hot
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if(get_uMsg(params) == WM_COMMAND && id == LOWORD(get_wParam(params))) {
				get_lResult(params) = (get_derived(params).*func)(HIWORD(get_wParam(params)), LOWORD(get_wParam(params)), (HWND)get_lParam(params), get_bHandled(params));
				return true;
			}
			return false;
		}
	};

	template< WORD id, command_handler_type func >
	struct command_id_handler_not
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if(get_uMsg(params) == WM_COMMAND && id == LOWORD(get_wParam(params))) {
				get_lResult(params) = (get_derived(params).*func)(HIWORD(get_wParam(params)), LOWORD(get_wParam(params)), (HWND)get_lParam(params), get_bHandled(params));
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_code_handler
	//
	template< WORD code, command_handler_type func >
	struct command_code_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if(get_uMsg(params) == WM_COMMAND && id == HIWORD(get_wParam(params))) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)(HIWORD(get_wParam(params)), LOWORD(get_wParam(params)), (HWND)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_range_handler
	//
	template< WORD idFirst, WORD idLast, command_handler_type func >
	struct command_range_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if(get_uMsg(params) == WM_COMMAND && LOWORD(get_wParam(params)) >= idFirst && LOWORD(get_wParam(params)) <= idLast) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)(HIWORD(get_wParam(params)), LOWORD(get_wParam(params)), (HWND)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// command_range_code_handler
	//
	template< WORD idFirst, WORD idLast, WORD code, command_handler_type func >
	struct command_range_code_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if(get_uMsg(params) == WM_COMMAND && code == HIWORD(get_wParam(params)) && LOWORD(get_wParam(params)) >= idFirst && LOWORD(get_wParam(params)) <= idLast) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)(HIWORD(get_wParam(params)), LOWORD(get_wParam(params)), (HWND)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_handler
	//
	typedef LRESULT (Derived::*notify_handler_type)(int, LPNMHDR, BOOL&);

	template< UINT id, UINT cd, notify_handler_type func >
	struct notify_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == WM_NOTIFY && id == ((LPNMHDR)get_lParam(params))->idFrom && cd == ((LPNMHDR)get_lParam(params))->code) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)((int)get_wParam(params), (LPNMHDR)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_id_handler
	//
	template< UINT id, notify_handler_type func >
	struct notify_id_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == WM_NOTIFY && id == ((LPNMHDR)get_lParam(params))->idFrom) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)((int)get_wParam(params), (LPNMHDR)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_code_handler
	//
	template< UINT cd, notify_handler_type func >
	struct notify_code_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == WM_NOTIFY && cd == ((LPNMHDR)get_lParam(params))->code) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)((int)get_wParam(params), (LPNMHDR)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	template< UINT cd, notify_handler_type func >
	struct notify_code_handler_hot
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == WM_NOTIFY && cd == ((LPNMHDR)get_lParam(params))->code) {
				get_lResult(params) = (get_derived(params).*func)((int)get_wParam(params), (LPNMHDR)get_lParam(params), get_bHandled(params));
				return true;
			}
			return false;
		}
	};
	
	template< UINT cd, notify_handler_type func >
	struct notify_code_handler_not
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == WM_NOTIFY && cd == ((LPNMHDR)get_lParam(params))->code) {
				get_lResult(params) = (get_derived(params).*func)((int)get_wParam(params), (LPNMHDR)get_lParam(params), get_bHandled(params));
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_range_handler
	//
	template< UINT idFirst, UINT idLast, notify_handler_type func >
	struct notify_range_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == WM_NOTIFY && ((LPNMHDR)get_lParam(params))->idFrom >= idFirst && ((LPNMHDR)get_lParam(params))->idFrom <= idLast) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)((int)get_wParam(params), (LPNMHDR)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// notify_range_code_handler
	//
	template< UINT idFirst, UINT idLast, UINT cd, notify_handler_type func >
	struct notify_range_code_handler
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_uMsg(params) == WM_NOTIFY &&  cd == ((LPNMHDR)get_lParam(params))->code && ((LPNMHDR)get_lParam(params))->idFrom >= idFirst && ((LPNMHDR)get_lParam(params))->idFrom <= idLast) {
				get_bHandled(params) = TRUE;
				get_lResult(params) = (get_derived(params).*func)((int)get_wParam(params), (LPNMHDR)get_lParam(params), get_bHandled(params));
				if (get_bHandled(params))
					return true;
			}
			return false;
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// chain_msg_map
	//
	//  Note: to avoid ::type syntax
	//
	template< class theChainClassOrMessageMap >
	struct chain_msg_map_base
	{
		template< class MessageMap >
		struct chain_if_statement
		{
			static bool compile(process_window_message_params_t_reference params)
			{
				return get_derived(params).process_window_message<MessageMap>(get_hWnd(params), get_uMsg(params), get_wParam(params), get_lParam(params), get_lResult(params), get_dwMsgMapID(params));
			}

			typedef chain_if_statement type;
		};

		template< class theChainClass >
		struct chain_if_statement_Atl
		{
			static bool compile(process_window_message_params_t_reference params)
			{
				if (get_derived(params).theChainClass::ProcessWindowMessage(get_hWnd(params), get_uMsg(params), get_wParam(params), get_lParam(params), get_lResult(params), get_dwMsgMapID(params)))
					return true;
				return false;
			}

			typedef chain_if_statement_Atl type;
		};

		typedef typename boost::mpl::eval_if<boost::mpl::is_sequence<theChainClassOrMessageMap>,
			chain_if_statement<boost::mpl::_>,
			chain_if_statement_Atl<boost::mpl::_>
		>::type if_statement;

		typedef typename boost::mpl::apply<if_statement, theChainClassOrMessageMap>::type type;
	};

	template< class theChainClassOrMessageMap >
	struct chain_msg_map : chain_msg_map_base<theChainClassOrMessageMap>::type
	{ };

	///////////////////////////////////////////////////////////////////////////////
	// alt_msg_map
	//
	template< DWORD msgMapID, class MessageMap >
	struct alt_msg_map
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			if (get_dwMsgMapID(params) == msgMapID) {
				return get_derived(params).process_window_message<MessageMap>(get_hWnd(params), get_uMsg(params), get_wParam(params), get_lParam(params), get_lResult(params), get_dwMsgMapID(params));
			}
			else
				return false; // pass to other alternates
		}
	};

	///////////////////////////////////////////////////////////////////////////////
	// cracked_handlers
	//

//#	include "detail/cracked_handlers.ipp"
	template<
		void (Derived::*func)()
	>
	struct msg_wm_close
	{
		static bool compile(process_window_message_params_t_reference params)
		{
			BOOST_STATIC_ASSERT(has_cracked_handlers);

			if (get_uMsg(params) == WM_CLOSE) {
				get_derived(params).set_msg_handled(true);
				(get_derived(params).*func)();
				get_lResult(params) = 0;
				if (get_derived(params).is_msg_handled())
					return true;
			}
			return false;
		}
	};


}; // struct message_processor

} // namespace ketchup

///////////////////////////////////////////////////////////////////////////////
// KETCHUP_CHAIN_MSG_MAP for ATL's BEGIN_MSG_MAP
// 
#define KETCHUP_CHAIN_MSG_MAP(MessageMap) {\
	if (process_window_message<MessageMap>(hWnd, uMsg, wParam, lParam, lResult, dwMsgMapID))\
		return TRUE;\
}

#endif
